#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#ifdef FOR_R
# include <R.h>
#else
#define warning(args...) fprintf (stderr, args)
#endif

/* #include <R_ext/Utils.h> */
/* void warning(const char * format, ...); */

/* #define  DEBUG */

int careful=1;

extern double unit_arcone(const int, double *, double *,  
			int, double, double);

extern void ar_free_mem();

#include "linalg.h"
#include "PSC.h"
  
decomp PSC_decomp(const int n, 
                  double V[n][n], double At[n][n], 
		  const double mu[n], 
                  int p, 
		  int log2G, double U, double S, int conesonly){

#ifdef DEBUG
  fprintf(stderr, "p: %d\n", p);
#endif

#ifdef FOR_R  
  R_CheckUserInterrupt();
#endif

  if(p>=n-1) {
    /* Call on unitarcone to do the work */

    double Sigma[n][n], xmu[n];
    int i,j;

    for(i=0; i<n; i++) {
      xmu[i] = vdot(n, At[i], mu);
      for(j=0; j<=i; j++) 
	Sigma[i][j] = vdot(n, At[i], At[j]);
    }

    double rho[n-1], sd[n];
    for(i=0; i<n; i++) {
      sd[i] = sqrt(Sigma[i][i]);
      xmu[i] /= sd[i];
    }
    for(i=0; i<n-1; i++) rho[i] = Sigma[i+1][i]/(sd[i]*sd[i+1]);

    
    if(careful) {
      double I[n][n], mnorm=0;

      for(i=0; i<n; i++) 
	for(j=0; j<n; j++) {
	  I[i][j] = vdot(n, V[i], At[j]);
	  mnorm += fabs(I[i][j]-(i==j));
	}
      
      double tau[n], alpha[n-1], psi[n-1], var[n], sig[n];

      for(i=0; i<n-1; i++) 
	psi[i] = vdot(n, V[i], V[i+1]);
      
      tau[n-1] = 1;
      
      for(i=n-2; i>=0; i--) {
	alpha[i] = -psi[i]/tau[i+1];
	tau[i] = 1-tau[i+1]*alpha[i]*alpha[i];
      }
      
      var[0] = 1/tau[0];
      sig[0] = sqrt(var[0]);
      for(i=1; i<n; i++) {
	var[i] = alpha[i-1]*alpha[i-1]*var[i-1]+(1/tau[i]);
	sig[i] = sqrt(var[i]);
	alpha[i-1] *= sig[i-1]/sig[i];
      }
      
      double vnorm=0, rhonorm=0;
      for(i=0; i<n; i++) vnorm += fabs(Sigma[i][i]-var[i])/Sigma[i][i];
      for(i=0; i<n-1; i++) rhonorm += fabs(rho[i]-alpha[i]);

#ifdef DEBUG      
      fprintf(stderr, 
	      "Mnorm: %g, vnorm: %g, rhonorm: %g\n", 
	      mnorm, vnorm, rhonorm);
#endif
      if(mnorm>1e-10) {
	warning("V may not be inverse of  A: mnorm= %g, vnorm= %g, rhonorm= %g\n", 
		mnorm, vnorm, rhonorm);
#ifdef DEBUG	
	for(i=0; i<n; i++) {
	  for(j=0; j<n; j++)
	    fprintf(stderr, "%g ", I[i][j]);
	  fprintf(stderr, "\n");
	}
#endif
      }
    }
    


    decomp val;
    if(!conesonly)
      val.val = unit_arcone(n, xmu, rho, log2G, U, S);
    else
      val.val = 0;
    val.ncone = 1;

#ifdef DEBUG
    fprintf(stderr, "Val: %g\n", val.val);
#endif
    return val;
  }
  
  /* OK, p<n-1 and so we now proceed recursively */

  int i,j;
  double Vnew[n][n], Atnew[n][n], dots[n-p];

  for(i=0; i<p; i++) {
    vcopy(Vnew[i], n, V[i]);
    vcopy(Atnew[i], n, At[i]);
  }

  for(i=0; i<n-p; i++) {
    dots[i] = vdot(n, V[p-1], V[p+i]);
    if (fabs(dots[i])*sqrt(vnorm2(n, At[p+i]))<1e-12) dots[i] = 0;
  }

#ifdef DEBUG
  fprintf(stderr, "Dots: ");
  for(i=0; i<n-p; i++)
    fprintf(stderr, "%g ", dots[i]);
  fprintf(stderr, "\n");
#endif  

  int index[n-p];
  for(i=0; i<n-p; i++) index[i] = i;
  int dcompare(const void *i1, const void *i2) {
    double x = dots[*(int *)i1], y=dots[*(int *)i2];
    if(x==y) return 0;
    return x<y ? 1 : -1;
  }
#ifdef DEBUG
  fprintf(stderr, "About to qsort\n");
#endif
  qsort(index, n-p, sizeof(int), dcompare);
#ifdef DEBUG
  fprintf(stderr, "Finished qsort\n");
#endif

  if((dots[index[0]]==0) && (dots[index[n-p-1]]==0)) 
    return PSC_decomp(n, V, At, mu, p+1, log2G, U, S, conesonly);

  double odots[n-p];
  double oV[n-p][n];
  double oAt[n-p][n];

  decomp val;
  val.ncone = 0;
  val.val = 0;

  void do_plus() {
    vcopy(Atnew[p], n, V[p-1]);
    vsub(Atnew[p], n, Atnew[p-1]);
    if(p>1)
      vaddm(Atnew[p], n, Atnew[p-2], -vdot(n, Vnew[p-1], Vnew[p-2]));
    for(i=0; (i<n-p) && (odots[i]>0); i++) {
      vcopy(Vnew[p], n, oV[i]);
      int col=p+1;
      for(j=0; j<n-p; j++) {
	if (j==i) continue;
	vcopym(Vnew[col], n, oV[j], j<i? -odots[i]: odots[i]);
	vaddm(Vnew[col], n, oV[i], j<i? odots[j]: -odots[j]);
	vscale(Vnew[col], n, 1/sqrt(vnorm2(n, Vnew[col])));
	vcopy(Atnew[col], n, oAt[j]);
	vscale(Atnew[col], n, 1/vdot(n, Atnew[col], Vnew[col]));
	col++;
      }
      vscale(Atnew[p], n, 1/vdot(n, Atnew[p], Vnew[p]));
      decomp tmp;
      tmp = PSC_decomp(n, Vnew, Atnew, mu, p+1, log2G, U, S, conesonly);
      val.ncone += tmp.ncone;
      if (i%2)
	val.val -= tmp.val;
      else
	val.val += tmp.val;
    }
  }


  for(i=0; i<n-p; i++) {
    odots[i] = dots[index[i]];
    vcopy(oV[i], n, V[p+index[i]]);
    vcopy(oAt[i], n, At[p+index[i]]);
  }
#ifdef DEBUG
  fprintf(stderr, "Dots: ");
  for(i=0; i<n-p; i++)
    fprintf(stderr, "%g ", odots[i]);
  fprintf(stderr, "\n");
#endif
    
  do_plus();
  for(i=0; i<n-p; i++) {
    odots[n-p-1-i] = -dots[index[i]];
    vcopy(oV[n-p-1-i], n, V[p+index[i]]);
    vcopy(oAt[n-p-1-i], n, At[p+index[i]]);
  }
#ifdef DEBUG
  fprintf(stderr, "Dots: ");
  for(i=0; i<n-p; i++)
    fprintf(stderr, "%g ", odots[i]);
  fprintf(stderr, "\n");
#endif
    
  do_plus();
#ifdef DEBUG
  fprintf(stderr, "Val: %g\n", val.val);
#endif
  return val;
}

#ifdef FOR_R
void PSC_decomp_R(const int *np, 
		  double V[*np][*np], 
		  double At[*np][*np], 
		  const double mu[*np], 
		  int *p, 
		  int *log2G, double *U, double *S,
		  int *conesonly,
		  int *ncone,
		  double *value) {
  decomp val;
  val = PSC_decomp(*np, V, At, mu, *p, *log2G, *U, *S, *conesonly);
  *ncone = val.ncone;
  *value  = val.val;
  ar_free_mem();
}
#endif
