#ifdef FOR_R
# include <R.h>
# include <Rmath.h>
#else
# include <stdlib.h>
# include <math.h>
#endif

#include "cft.h"

#ifdef FOR_R
cftfunction_t *cftp;
#else
cftfunction_t *cftp=&cft;
#endif



/* #define TROUBLE */
/* #define DEBUG 1 */

extern double exp2(double);
extern double log2(double);

static inline int max(const int x, const int y) {
  return x>y?x:y;
}

static inline int min(const int x, const int y) {
  return x<y?x:y;
}

static inline double sqr(double x) {
  return x*x;
}

static inline double dsnorm(const double x) {
#ifdef USING_R
  return dnorm(x, 0, 1, 0);
#else
  return exp(-x*x/2)/sqrt(2*M_PI);
#endif
}

#ifdef USING_R
static inline double psnorm(const double x) {
  return pnorm(x, 0, 1, 1, 0);
}
#else
# define Calloc(x,y) (y*)calloc((unsigned)(x),sizeof(y))
# define Free(x) free(x)
#endif





void ar_free_mem();


typedef double zetas[4];

static int myerf_nm=0;
static double *myerfs, *myexps;
static int myerf_lowm, myerf_highm, zeta_lowm, zeta_highm;
static int xi_lowm, xi_highm;

static void setup_myerfs(double mu, double sigma, double U) {
  zeta_lowm = ceil(-mu-U*sigma);
  zeta_highm = ceil(-mu+U*sigma);
  myerf_lowm = zeta_lowm-1;
  myerf_highm = zeta_highm;
  xi_lowm = zeta_lowm-2;
  xi_highm = zeta_highm+1;
  int nm = myerf_highm-myerf_lowm+1;
#ifdef DEBUG
  printf("%d %d %d\n", myerf_lowm, myerf_highm, nm);
#endif
 
  if(nm>myerf_nm) {
    if(myerf_nm) {
      Free(myerfs);
      Free(myexps);
    } else myerf_nm = 1;
    while(myerf_nm<nm) myerf_nm *=2;
    myerfs = Calloc(myerf_nm, double);
    myexps = Calloc(myerf_nm, double);
  }
  
  int i;
  for(i=0; i<nm; i++) {
    int m = i+myerf_lowm;
#ifdef USING_R
    myerfs[i] = 2*psnorm((mu+m)/sigma)-1;
#else
    myerfs[i] = erf((mu+m)/sigma/sqrt(2));
#endif    
    myexps[i] = exp(-0.5*sqr((mu+m)/sigma));  }
}


/* 
 * Based on current mu and sigma, computes
 * erf( (mu+m)/(sigma\sqrt{2} )
 * which is 2Phi((mu+m)/sigma)-1
 */
double myerf(int m) {
  if(m<myerf_lowm) {
    /* printf("Warning: m too low in myerf (%d)\n", m); */
    return(-1);
  }
  if(m>myerf_highm) {
    /* printf("Warning: m too high in myerf (%d)\n", m); */
    return(1);
  }
  return myerfs[m-myerf_lowm];
}


double myexp(int m) {
  if(m<myerf_lowm) {
    /* printf("Warning: m too low in myexp (%d)\n", m); */
    return(0);
  }
  if(m>myerf_highm) {
    /* printf("Warning: m too high in myexp (%d)\n", m); */
    return(0);
  }
  return myexps[m-myerf_lowm];
}


#include "zeta.cfrag"
#include "zetaseries.cfrag"

void zeta(double mu, double sigma, int m, zetas z) {
  if(sigma>4) 
    Czetaseries((mu+m)/sigma, 1/sigma, z);
  else
    Czeta(mu, sigma, m, z);
}

void zetastar(double mu, double sigma, int m, zetas z) {
  if(sigma>4) 
    Czetasseries((mu+m)/sigma, 1/sigma, z);
  else
    Czetas(mu, sigma, m, z);
}

void zetastar2(double mu, double sigma, int m, zetas z) {
  if(sigma>4) 
    Czetas2series((mu+m)/sigma, 1/sigma, z);
  else
    Czetas2(mu, sigma, m, z);
}

#include "xi.cfrag"
#include "xiseries.cfrag"

double xi(double mu, double sigma, int m) {
  if(sigma>4) 
    return Cxiseries((mu+m)/sigma, 1/sigma);
  else
    return Cxi(mu, sigma, m);
}

Rfuncapprox linfilt2(Rfuncapprox psi, double rho, double sigma, 
		    double start, double U, double *psilf) {
  const int G=psi.n;
  double deltastar, mustar, sigmastar;
  if(rho>=0) {
    deltastar = rho*psi.delta;
    mustar = (start-rho*psi.start)/deltastar;
  } else {
    /* 
     * We handle negative rho by changing the sign of the x-values and
     * rho. We have to recalculate the grid starting point because the
     * order is reversed by changing the sign. After filtering we have
     * to reorder the results again
     */
    deltastar = -rho*psi.delta;
    double start2 = -(start+(G-1)*deltastar);
    mustar = (start2+rho*psi.start)/deltastar;
  }
  sigmastar = sigma/deltastar;
  if(sigmastar<=1e-12) sigmastar = 1e-12;
	
  setup_myerfs(mustar, sigmastar, U);
      
  int m, i;
  double zetam[4];
  for(m=xi_lowm; m<=xi_highm; m++) {
    /* Do ordinary interior interval contributions first */
    int startk = max(3,-m);
    int endk = min(G-3,G-m);
    if (startk>=endk) continue;
    double xim = xi(mustar, sigmastar, m);
    int k;
    for(k=startk; k<endk; k++) 
      psilf[m+k] += xim*psi.vals[k];
  }
  /*
   *  Handle non-edge-interval contributions from points near the
   *  edges. The actual contribution is from old psi[k-1+i] to new
   *  psi[m+k]. Hence, need 1<=k<G-2 and 0<=m+k<G and (0<=k-1+i<3 or
   *  G-3<=k-1+i<G)
   */
  for(m=zeta_lowm; m<=zeta_highm; m++) {
    int zetaevaluated = 0;
    for(i=0; i<4; i++) {
      int edge;
      for(edge=0; edge<2; edge++) {
	int startk, endk;
	if(edge==0) {
	  startk = max(1,-m); /* Also needs to be >= 1-i but automatic */
	  endk = min(3+1-i, min(G-2,G-m));
	} else {
	  startk = max(endk, G-3+1-i);
	  endk = min(G+1-i,min(G-2, G-m));
	}
	if(startk<endk) {
	  if (!zetaevaluated) {
	    zeta(mustar, sigmastar, m, zetam);
	    zetaevaluated = 1;
	  }
	  int k;
	  for(k=startk; k<endk; k++) 
	    psilf[m+k] += zetam[i]*psi.vals[k-1+i];
	} else 
	  endk = startk;
      }
    }
  }
  /* Now do left-edge-interval contributions */
  for(m=max(zeta_lowm, 0); m<=min(zeta_highm, G-1); m++) {
    zetastar(mustar, sigmastar, m, zetam);
    for(i=0; i<4; i++) 
      psilf[m] += zetam[i] * psi.vals[i];
  }
  /* and right-edge-interval contributions */
  for(m=max(zeta_lowm, -(G-2)); m<=min(zeta_highm, (G-1)-(G-2)); m++) {
    zetastar2(mustar, sigmastar, m, zetam);
    for(i=0; i<4; i++) 
      psilf[m+(G-2)] += zetam[i] * psi.vals[G-4+i];
  }
  int l;
  if(rho>=0)
    for(l=0; l<G; l++) 
      psi.vals[l] = psilf[l]/(rho*sigmastar);
  else
    for(l=0; l<G; l++) 
      psi.vals[G-1-l] = -psilf[l]/(rho*sigmastar);
  psi.start = start;
  psi.delta = deltastar;
  return psi;
}

void linfilt2_R(double *x, int *np, double *startp, double *deltap, 
	       double *rhop, double *sigmap, double *newstartp, double *Up) {
  Rfuncapprox f;
  
  f.vals = x;
  f.n = *np;
  f.nmax = f.n;
  f.start = *startp;
  f.delta = *deltap;
  
  double *psilp = Calloc(f.n, double);
  f = linfilt2(f, *rhop, *sigmap, *newstartp, *Up, psilp);
  ar_free_mem();
}

static double doint(funcapprox psi) {
  complex integral = 0;
  complex *y=psi.vals;
  int i, G=psi.n;
  
  for(i=0; i<G; i++)
    integral += y[i];
  integral -= (2./3.)*(y[0]+y[G-1]);
  integral += (7./24.)*(y[1]+y[G-2]);
  integral -= (1./6.)*(y[2]+y[G-3]);
  integral += (1./24.)*(y[3]+y[G-4]);
  integral *= psi.delta;
  
  return creal(integral);

}

static void initpsi(funcapprox *psi, 
		    double start, double delta, int n, 
		    double factor) {
  int i;

  psi->start = start;
  psi->delta = delta;
  psi->n = n;
  for(i=0; i<n; i++) 
    psi->vals[i] = factor*dsnorm(start+i*delta);
}

static funcapprox psi = {0.,0.,0,0,NULL};
static int npsilf=0;
static double *psilf=NULL; /* Used only when linear filtering */

double unit_arcone(const int n, double *mu, double *rho,  
		 int log2G, double U, double S) {
  

#ifdef DEBUG
  printf("n is %d\n", n);
  printf("log2G is %d\n", log2G);
#endif

  if(mu[0] < -U) { 
    return 0; 
  }

  int i, G;
  double delta, deltas[n];

  G = exp2(log2G);
  delta = 2*U/G;

  double sigma[n-1];
  int g[n-1], gmax=1;
  deltas[n-1] = delta;
  for(i=n-2; i>=0; i--) {
    if(fabs(rho[i])<1e-12) {
      rho[i] = 0;
      deltas[i] = delta;
      g[i] = -1;
      continue;
    }
    sigma[i] = sqrt(1-sqr(rho[i]));
    if(sigma[i]==0) {
      g[i] = 0;
      deltas[i] = deltas[i+1];
      continue;
    }
    double ratio = fabs(rho[i])*(log2G-log2(sigma[i]))/(G*(1-sqr(rho[i])));
    if(ratio>S) {
      /* Try to use linear filtering */
      deltas[i] = deltas[i+1]/fabs(rho[i]);
      if(deltas[i]<=2*delta) {
	g[i] = 0;
	continue;
      }
      /* trouble. look forward to replace a linear filter by an FFT */
      int j, pos = i;
      double max = sigma[i];
      for(j=i+1; (j<n-1) && (g[j]==0); j++) {
	if (sigma[j]>max) {
	  max = sigma[j];
	  pos = j;
	}
      }
      /* Make stage pos be FFT and then redo calc from there */
      i = pos;
    }
    /* Use FFT */
    g[i] = exp2(-ceil(log2(sigma[i])));
    if(g[i]>gmax) gmax = g[i];
    deltas[i] = delta;
  }


#ifdef DEBUG
  printf("gmax is %d, G is %d\n", gmax, G);
#endif

  if(psi.nmax<gmax*G) {
    if(psi.nmax) Free(psi.vals);
    psi.nmax = gmax*G;
    psi.vals = Calloc(psi.nmax, complex);
  }

  double zstart;

  zstart = -mu[0];
  if (zstart < -U) zstart = -U;
  
  initpsi(&psi, zstart, deltas[0], G, 1);

  int j;

#ifdef TROUBLE
  FILE *ftrouble = fopen("trouble.dat", "w");
#endif

  for(j=1; j<n; j++) {
    double sigmaj = sigma[j-1];
    double rhoj = rho[j-1];
    int gj = g[j-1];
    

    zstart = -mu[j];
    if (zstart < -U) zstart = -U;

    if (rhoj==0) {
      /* No correlation. Integrate old bit and start again */
      double current = doint(psi);
      initpsi(&psi, zstart, deltas[j], G, current);
    } else if(gj==0) {
      /* Linear filtering */
      /* First time in here, need to allocate the storage for the
	 computation */
      if (npsilf<2*G) {
	if(npsilf) Free(psilf);
	npsilf = 2*G;
	psilf = Calloc(npsilf, double);
      }
      Rfuncapprox Rpsi;
      Rpsi.n = G;
      Rpsi.nmax = G;
      Rpsi.start = psi.start;
      Rpsi.delta = psi.delta;
      Rpsi.vals = psilf;
      for(i=0; i<G; i++) {
	Rpsi.vals[i] = creal(psi.vals[i]);
	psilf[G+i] = 0;
      }
      Rpsi = linfilt2(Rpsi, rhoj, sigmaj, zstart, U, psilf+G);
      psi.start = Rpsi.start;
      psi.delta = Rpsi.delta;
      for(i=0; i<G; i++) 
	psi.vals[i] = Rpsi.vals[i];
#ifdef TROUBLE
      for(i=0; i<G; i++) 
	fprintf(ftrouble, "%20e\n", Rpsi.vals[i]);
#endif
    } else {
      /* FFT */
      /* Compute old Psi+ on grid centered at 0 with spacing
	 \delta\rho */
      psi = (*cftp)(psi, -rhoj*U/sigmaj, 2*rhoj*U/sigmaj/(gj*G), gj, 0, 1);
      /* Compute new Psi on grid centered at 0 with spacing \delta */
      psi.start /= rhoj;
      psi.delta /= rhoj;
      double z;
      for(i=0; i<psi.n; i++) {
	z = psi.start+i*psi.delta;
	psi.vals[i] *= exp(-sigmaj*sigmaj*z*z/2);
      }
      
      /* Compute new psi(+) on grid starting at max(-\mu_j, -U) with
       spacing \delta */
      zstart = -mu[j];
      if (zstart < -U) zstart = -U;
      
      psi = (*cftp)(psi, zstart, deltas[j], 1, 1, 1);
      psi.n = G;
    }
  }
  
#ifdef TROUBLE
  fclose(ftrouble);
#endif

  return doint(psi);
}

void unit_arcone_R(const int *npointer, double *mu, double *rho,  
		 int *log2Gp, double *Up, double *Sp,
		 double *val) {
  *val = unit_arcone(*npointer, mu, rho, *log2Gp, *Up, *Sp);
  ar_free_mem();
}

void ar_free_mem() {
  if(myerf_nm) {
    Free(myerfs);
    Free(myexps);
    myerf_nm = 0;
  }
  if(npsilf) {
    Free(psilf);
    npsilf = 0;
  }
  if(psi.nmax) {
    Free(psi.vals);
    psi.nmax = 0;
  }
      
}
